""" 
io.py contains the classes used to generate the Direct Commands for the EV3
brick. Direct Commands are grouped by Brick, Motor, and Sensor.

Author: Eduardo Nigro
    rev 0.0.1
    2021-10-24
"""
import struct
import numpy as np


class OpCode:

    """ Class containing EV3 op code definitions."""

    opUI_READ = 0x81
    opUI_WRITE = 0x82
    opSOUND = 0x94
    opINPUT_DEVICE_LIST = 0x98
    opINPUT_DEVICE = 0x99
    opOUTPUT_STOP = 0xA3
    opOUTPUT_POWER = 0xA4
    opOUTPUT_SPEED = 0xA5
    opOUTPUT_START = 0xA6
    opOUTPUT_READ = 0xA8
    opOUTPUT_CLR_COUNT = 0xB2
    opOUTPUT_GET_COUNT = 0xB3
    opCOM_GET = 0xD3
    opOUTPUT_STEP_POWER = 0xAC
    opOUTPUT_TIME_POWER = 0xAD
    opOUTPUT_STEP_SPEED = 0xAE
    opOUTPUT_TIME_SPEED = 0xAF


class SubCodeComGet:

    """ UIWriteSubCode - EV3 UI_WRITE subcode definition."""

    GET_PIN = 0x05
    GET_ID = 0x0C
    GET_BRICKNAME = 0x0D
    GET_NETWORK = 0x0E


class SubCodeInputDevice:

    """ OpCode - EV3 input subcode definition."""

    GET_FORMAT = 0x02
    GET_TYPEMODE = 0x05
    GET_SYMBOL = 0x06
    GET_NAME = 0x15
    GET_MODENAME = 0x16
    CLR_CHANGES = 0x1A
    READY_PCT = 0x1B
    READY_RAW = 0x1C
    READY_SI = 0x1D
    GET_MINMAX = 0x1E


class SubCodeUIRead:

    """ UIReadSubCode - EV3 UI_READ subcode definition. """

    GET_OS_VERS = 0x03
    GET_HW_VERS = 0x09
    GET_FW_VERS = 0x0A
    GET_FW_BUILD = 0x0B
    GET_OS_BUILD = 0x0C
    GET_LBATT = 0x12


class SubCodeUIWrite:

    """ UIWriteSubCode - EV3 UI_WRITE subcode definition. """

    LED = 0x1B


class DataFlag:

    """ DataFlag - EV3 DataFlag definition. """

    PRIMPAR_SHORT = 0x00
    PRIMPAR_LONG = 0x80
    PRIMPAR_CONST = 0x00
    PRIMPAR_VARIABLE = 0x40
    PRIMPAR_LOCAL = 0x00
    PRIMPAR_GLOBAL = 0x20
    PRIMPAR_HANDLE = 0x10
    PRIMPAR_ADD = 0x08
    PRIMPAR_INDEX = 0x1F
    PRIMPAR_CONST_SIGN = 0x20
    PRIMPAR_VALUE = 0x3F
    PRIMPAR_BYTES = 0x07
    PRIMPAR_STRING_OLD = 0
    PRIMPAR_1_BYTES = 1
    PRIMPAR_2_BYTES = 2
    PRIMPAR_4_BYTES = 3
    PRIMPAR_STRING = 4


class DeviceType:

    """ Device (sensor) type identifyer """

    TYPE_EV3_TOUCH = 16
    TYPE_EV3_COLOR = 29
    TYPE_EV3_ULTRASONIC = 30
    TYPE_EV3_GYRO = 32
    TYPE_EV3_IR = 33
    TYPE_THIRD_PARTY_START = 50
    TYPE_THIRD_PARTY_END = 99
    TYPE_IIC_UNKNOWN = 100
    TYPE_TERMINAL = 124
    TYPE_UNKNOWN = 125
    TYPE_NONE  = 126
    TYPE_ERROR = 127


class ByteCode:

    """ ByteCode - EV3 byte code generation. """

    def __init__(self):
        self.numLocal = 0
        self.numGlobal = 0
        self.bCode = []

    # BRICK 
    def get_id(self, HARDWARE):
        """ opCOM_GET(GET_ID, (uint8)HARDWARE, (uint8)LENGTH, *(uint8)STRING) """
        self.numLocal = 4
        self.numGlobal = 0
        LENGTH = 16
        self.bCode = b''.join([
            OpCode.opCOM_GET.to_bytes(1, 'little'),
            SubCodeComGet.GET_ID.to_bytes(1, 'little'),
            HARDWARE.to_bytes(1, 'little'),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def get_brickname(self):
        """ opCOM_GET(GET_BRICKNAME, (uint8)LENGTH, *(uint8)NAME) """
        self.numLocal = 2
        self.numGlobal = 0
        LENGTH = 8
        self.bCode = b''.join([
            OpCode.opCOM_GET.to_bytes(1, 'little'),
            SubCodeComGet.GET_BRICKNAME.to_bytes(1, 'little'),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def read_osversion(self):
        """ opUI_READ(GET_OS_VERS, (uint8)LENGTH, *(uint8)DESTINATION) """
        self.numLocal = 4
        self.numGlobal = 0
        LENGTH = 16
        self.bCode = b''.join([
            OpCode.opUI_READ.to_bytes(1, 'little'),
            SubCodeUIRead.GET_OS_VERS.to_bytes(1, 'little'),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def read_hardwareversion(self):
        """ opUI_READ(GET_HW_VERS, (uint8)LENGTH, *(uint8)DESTINATION) """
        self.numLocal = 2
        self.numGlobal = 0
        LENGTH = 8
        self.bCode = b''.join([
            OpCode.opUI_READ.to_bytes(1, 'little'),
            SubCodeUIRead.GET_HW_VERS.to_bytes(1, 'little'),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def read_firmwareversion(self):
        """ opUI_READ(GET_FW_VERS, (uint8)LENGTH, *(uint8)DESTINATION) """
        self.numLocal = 2
        self.numGlobal = 0
        LENGTH = 8
        self.bCode = b''.join([
            OpCode.opUI_READ.to_bytes(1, 'little'),
            SubCodeUIRead.GET_FW_VERS.to_bytes(1, 'little'),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def read_firmwarebuild(self):
        """ opUI_READ(GET_FW_BUILD, (uint8)LENGTH, *(uint8)DESTINATION) """
        self.numLocal = 4
        self.numGlobal = 0
        LENGTH = 16
        self.bCode = b''.join([
            OpCode.opUI_READ.to_bytes(1, 'little'),
            SubCodeUIRead.GET_FW_BUILD.to_bytes(1, 'little'),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def read_osbuild(self):
        """ opUI_READ(GET_OS_BUILD, (uint8)LENGTH, *(uint8)DESTINATION) """
        self.numLocal = 4
        self.numGlobal = 0
        LENGTH = 16
        self.bCode = b''.join([
            OpCode.opUI_READ.to_bytes(1, 'little'),
            SubCodeUIRead.GET_OS_BUILD.to_bytes(1, 'little'),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def read_batterylevel(self):
        """ opUI_READ(GET_LBATT, *(uint8)PCT) """
        self.numLocal = 0
        self.numGlobal = 1
        self.bCode = b''.join([
            OpCode.opUI_READ.to_bytes(1, 'little'),
            SubCodeUIRead.GET_LBATT.to_bytes(1, 'little'),
            struct.pack('<B', 96)
        ])

    def write_LED(self, PATTERN):
        """ opUI_WRITE(LED, (uint8)PATTERN) """
        self.numLocal = 0
        self.numGlobal = 0
        self.bCode = b''.join([
            OpCode.opUI_WRITE.to_bytes(1, 'little'),
            SubCodeUIWrite.LED.to_bytes(1, 'little'),
            PATTERN.to_bytes(1, 'little')
        ])

    def play_tone(self, volume, frequency, duration):
        """ opSOUND(TONE, (uint8)VOLUME, (int16)FREQUENCY, (int16)DURATION) """
        self.numLocal = 0
        self.numGlobal = 0
        TONE = 1
        VOLUME = LC1(int(volume))
        FREQUENCY = LC2(int(frequency))
        DURATION = LC2(int(duration * 1000))
        duration = int(duration * 1000)
        self.bCode = b''.join([
            OpCode.opSOUND.to_bytes(1, 'little'),
            struct.pack('<B', TONE),
            VOLUME,
            FREQUENCY,
            DURATION
        ])

    # MOTOR INTERFACE
    def output_stop(self, layer, port, brake):
        """ opOUTPUT_STOP((uint8)LAYER, (uint8)NOS, (uint8)BRAKE) """
        self.numLocal = 0
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = get_outputportid(port)
        BRAKE = get_outputbrakevalue(brake)
        self.bCode = b''.join([
            OpCode.opOUTPUT_STOP.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', BRAKE)
        ])

    def output_power(self, layer, port, power):
        """ opOUTPUT_POWER((uint8)LAYER, (uint8)NOS, (uint8)POWER) """
        self.numLocal = 0
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = get_outputportid(port)
        POWER = LC1(int(power))
        self.bCode = b''.join([
            OpCode.opOUTPUT_POWER.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            POWER
        ])

    def output_speed(self, layer, port, speed):
        """ opOUTPUT_SPEED((uint8)LAYER, (uint8)NOS, (uint8)SPEED) """
        self.numLocal = 0
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = get_outputportid(port)
        SPEED = LC1(int(speed))
        self.bCode = b''.join([
            OpCode.opOUTPUT_SPEED.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            SPEED
        ])

    def output_start(self, layer, port):
        """ opOUTPUT_START((uint8)LAYER, (uint8)NOS) """
        self.numLocal = 0
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = get_outputportid(port)
        self.bCode = b''.join([
            OpCode.opOUTPUT_START.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS)
        ])

    def output_clearcount(self, layer, port):
        """ opOUTPUT_CLR_COUNT((uint8)LAYER, (uint8)NOS) """
        self.numLocal = 0
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = get_outputportid(port)
        self.bCode = b''.join([
            OpCode.opOUTPUT_CLR_COUNT.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS)
        ])

    def output_getcount(self, layer, port):
        """ opOUTPUT_GET_COUNT((uint8)LAYER, (uint8)NOS, *(int32)TACHO) """
        self.numLocal = 1
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = get_outputportnum(port)
        self.bCode = b''.join([
            OpCode.opOUTPUT_GET_COUNT.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', 96)
        ])

    # SENSOR INTERFACE
    def input_devicelist(self):
        """ opINPUT_DEVICE_LIST((uint8)LENGTH, *(uint8)ARRAY, *(uint8)CHANGED) """
        self.numLocal = 2
        self.numGlobal = 0
        LENGTH = 8
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE_LIST.to_bytes(1, 'little'),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96),
            struct.pack('<B', 100)
        ])

    def input_deviceformat(self, layer, nos):
        """ opINPUT_DEVICE(GET_FORMAT, (uint8)LAYER, (uint8)NOS, 
            *(uint8)DATASETS, *(uint8)FORMAT, *(uint8)MODES, *(uint8)VIEW) """
        self.numLocal = 1
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = nos - 1
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE.to_bytes(1, 'little'),
            SubCodeInputDevice.GET_FORMAT.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', 96),
            struct.pack('<B', 97),
            struct.pack('<B', 98),
            struct.pack('<B', 99)
        ])

    def input_devicetypemode(self, layer, nos):
        """ opINPUT_DEVICE(GET_TYPEMODE, (uint8)LAYER, (uint8)NOS, *(uint8)TYPE, *(uint8)MODE) """
        self.numLocal = 0
        self.numGlobal = 2
        LAYER = layer - 1
        NOS = nos - 1
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE.to_bytes(1, 'little'),
            SubCodeInputDevice.GET_TYPEMODE.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', 96),
            struct.pack('<B', 97)
        ])

    def input_devicesymbol(self, layer, nos):
        """ opINPUT_DEVICE(GET_SYMBOL, (uint8)LAYER, (uint8)NOS, (uint8)LENGTH, *(uint8)DESTINATION) """
        self.numLocal = 4
        self.numGlobal = 0
        LENGTH = 16
        LAYER = layer - 1
        NOS = nos - 1
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE.to_bytes(1, 'little'),
            SubCodeInputDevice.GET_SYMBOL.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def input_devicename(self, layer, nos):
        """ opINPUT_DEVICE(GET_NAME, (uint8)LAYER, (uint8)NOS, (uint8)LENGTH, *(uint8)DESTINATION) """
        self.numLocal = 4
        self.numGlobal = 0
        LENGTH = 16
        LAYER = layer - 1
        NOS = nos - 1
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE.to_bytes(1, 'little'),
            SubCodeInputDevice.GET_NAME.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def input_devicemodename(self, layer, nos, mode):
        """ opINPUT_DEVICE(GET_MODENAME, (uint8)LAYER, (uint8)NOS, (uint8)MODE, (uint8)LENGTH, *(uint8)DESTINATION) """
        self.numLocal = 4
        self.numGlobal = 0
        LENGTH = 16
        LAYER = layer - 1
        NOS = nos - 1
        MODE = mode
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE.to_bytes(1, 'little'),
            SubCodeInputDevice.GET_MODENAME.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', MODE),
            struct.pack('<B', LENGTH),
            struct.pack('<B', 96)
        ])

    def input_deviceclrchanges(self, layer, nos):
        """ opINPUT_DEVICE(CLR_CHANGES, (uint8)LAYER, (uint8)NOS) """
        self.numLocal = 0
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = nos - 1
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE.to_bytes(1, 'little'),
            SubCodeInputDevice.CLR_CHANGES.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS)
        ])

    def input_devicereadySI(self, layer, nos, mode, values):
        """ opINPUT_DEVICE(READY_SI, (uint8)LAYER, (uint8)NOS, (uint8)TYPE, 
                (uint8)MODE, (uint8)VALUES, *(single32)VALUE1) """
        self.numLocal = 4
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = nos - 1
        TYPE = 0
        MODE = mode
        VALUES = values
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE.to_bytes(1, 'little'),
            SubCodeInputDevice.READY_SI.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', TYPE),
            struct.pack('<B', MODE),
            struct.pack('<B', VALUES)
        ])
        for i in range(VALUES):
            self.bCode += struct.pack('<B', 96+i*4)

    def input_devicereadyRAW(self, layer, nos, mode, values):
        """ opINPUT_DEVICE(READY_RAW, (uint8)LAYER, (uint8)NOS, (uint8)TYPE,
                (uint8)MODE, (uint8)VALUES, *(int32)VALUE1) """
        self.numLocal = 4
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = nos - 1
        TYPE = 0
        MODE = mode
        VALUES = values
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE.to_bytes(1, 'little'),
            SubCodeInputDevice.READY_RAW.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', TYPE),
            struct.pack('<B', MODE),
            struct.pack('<B', VALUES)
        ])
        for i in range(VALUES):
            self.bCode += struct.pack('<B', 96+i*4)

    def input_devicereadyPCT(self, layer, nos, mode, values):
        """ opINPUT_DEVICE(READY_PCT, (uint8)LAYER, (uint8)NOS, (uint8)TYPE,
                (uint8)MODE, (uint8)VALUES, *(uint8)VALUE1) """
        self.numLocal = 1
        self.numGlobal = 0
        LAYER = layer - 1
        NOS = nos - 1
        TYPE = 0
        MODE = mode
        VALUES = values
        self.bCode = b''.join([
            OpCode.opINPUT_DEVICE.to_bytes(1, 'little'),
            SubCodeInputDevice.READY_PCT.to_bytes(1, 'little'),
            struct.pack('<B', LAYER),
            struct.pack('<B', NOS),
            struct.pack('<B', TYPE),
            struct.pack('<B', MODE),
            struct.pack('<B', VALUES)
        ])
        for i in range(VALUES):
            self.bCode += struct.pack('<B', 96+i)


class DirectCommand:

    """ DirectCommand - EV3 direct command definition """

    # BRICK
    def get_brickid(hardware):
        BCode = ByteCode()
        BCode.get_id(hardware)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def get_brickname():
        BCode = ByteCode()
        BCode.get_brickname()
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_osversion():
        BCode = ByteCode()
        BCode.read_osversion()
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_hardwareversion():
        BCode = ByteCode()
        BCode.read_hardwareversion()
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_firmwareversion():
        BCode = ByteCode()
        BCode.read_firmwareversion()
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_firmwarebuild():
        BCode = ByteCode()
        BCode.read_firmwarebuild()
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_osbuild():
        BCode = ByteCode()
        BCode.read_osbuild()
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_batterylevel():
        BCode = ByteCode()
        BCode.read_batterylevel()
        cmd = constructCommand('Reply', BCode)
        return cmd

    def write_LED(pattern):
        BCode = ByteCode()
        BCode.write_LED(pattern)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def play_tone(volume, frequency, duration):
        BCode = ByteCode()
        BCode.play_tone(volume, frequency, duration)
        cmd = constructCommand('Reply', BCode)
        return cmd

    # MOTORS
    def stop_motor(layer, port, brake):
        BCode = ByteCode()
        BCode.output_stop(layer, port, brake)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def set_outputpower(layer, port, power):
        BCode = ByteCode()
        BCode.output_power(layer, port, power)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def set_outputspeed(layer, port, speed):
        BCode = ByteCode()
        BCode.output_speed(layer, port, speed)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def start_motor(layer, port):
        BCode = ByteCode()
        BCode.output_start(layer, port)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def clear_outputcount(layer, port):
        BCode = ByteCode()
        BCode.output_clearcount(layer, port)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def get_outputcount(layer, port):
        BCode = ByteCode()
        BCode.output_getcount(layer, port)
        cmd = constructCommand('Reply', BCode)
        return cmd

    # SENSORS
    def read_inputdevicelist():
        BCode = ByteCode()
        BCode.input_devicelist()
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_inputdeviceformat(layer, portnum):
        BCode = ByteCode()
        BCode.input_deviceformat(layer, portnum)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_inputdevicetypemode(layer, portnum):
        BCode = ByteCode()
        BCode.input_devicetypemode(layer, portnum)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_inputdevicesymbol(layer, portnum):
        BCode = ByteCode()
        BCode.input_devicesymbol(layer, portnum)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_inputdevicename(layer, portnum):
        BCode = ByteCode()
        BCode.input_devicename(layer, portnum)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_inputdevicemodename(layer, portnum, mode):
        BCode = ByteCode()
        BCode.input_devicemodename(layer, portnum, mode)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def clear_inputdevicechanges(layer, portnum):
        BCode = ByteCode()
        BCode.input_deviceclrchanges(layer, portnum)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_inputdevicereadySI(layer, portnum, mode, dataset):
        BCode = ByteCode()
        BCode.input_devicereadySI(layer, portnum, mode, dataset)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_inputdevicereadyRAW(layer, portnum, mode, dataset):
        BCode = ByteCode()
        BCode.input_devicereadyRAW(layer, portnum, mode, dataset)
        cmd = constructCommand('Reply', BCode)
        return cmd

    def read_inputdevicereadyPCT(layer, portnum, mode, dataset):
        BCode = ByteCode()
        BCode.input_devicereadyPCT(layer, portnum, mode, dataset)
        cmd = constructCommand('Reply', BCode)
        return cmd


# HELPER FUNCTIONS
def constructCommand(cmdtype, bcode):

    """ Assembles direct command using input byte code"""

    # Getting command type
    if cmdtype.lower() == 'reply':
        type = 0x00
    else:
        type = 0x80
    # Assigning comand pieces
    size = len(bcode.bCode)
    nl = bcode.numLocal
    nr = bcode.numGlobal
    nl = nl << 2
    nr1 = nr & 3
    nr2 = nr >> 2
    # Assembling direct command
    cmd = b''.join([
        struct.pack('<H', size + 5),
        struct.pack('<H', 0),
        type.to_bytes(1, 'little'),
        struct.pack('<B', nl + nr1),
        struct.pack('<B', nr2),
        bcode.bCode
    ])
    return cmd

def LC1(v):

    """ Parameter encoding - one byte to follow """

    b1 = DataFlag.PRIMPAR_LONG
    b2 = DataFlag.PRIMPAR_CONST
    b3 = DataFlag.PRIMPAR_1_BYTES
    result = b''.join([
        struct.pack('<B', ((b1 | b2) | b3)), 
        struct.pack('<B', np.array(v, dtype=np.int8).view(np.uint8))
        ])
    return result

def LC2(v):

    """ Parameter encoding - two bytes to follow """

    b1 = DataFlag.PRIMPAR_LONG
    b2 = DataFlag.PRIMPAR_CONST
    b3 = DataFlag.PRIMPAR_2_BYTES
    result = b''.join([
        struct.pack('<B', ((b1 | b2) | b3)), 
        struct.pack('<H', np.array(v, dtype=np.int16).view(np.uint16))
        ])
    return result

def LC4(v):

    """ Parameter encoding - four bytes to follow """

    b1 = DataFlag.PRIMPAR_LONG
    b2 = DataFlag.PRIMPAR_CONST
    b3 = DataFlag.PRIMPAR_4_BYTES
    result = b''.join([
        struct.pack('<B', ((b1 | b2) | b3)), 
        struct.pack('<I', np.array(v, dtype=np.int32).view(np.uint32))
        ])
    print([byte[0] for byte in list(struct.iter_unpack('B', result))])
    return result

def get_outputportid(port):

    """ Get the output port ID corresponding to the port letter """

    portid = {
        'a': 1,
        'b': 2,
        'c': 4,
        'd': 8
    }
    port = port.lower()
    if type(port) is not str:
        raise NameError('Invalid output port.')
    if port not in list(portid.keys()):
        raise NameError('Invalid output port.')
    return portid[port]

def get_outputportnum(port):

    """ Get the output port number corresponding to the port letter """

    portnum = {
        'a': 0,
        'b': 1,
        'c': 2,
        'd': 3
    }
    port = port.lower()
    if type(port) is not str:
        raise NameError('Invalid output port.')
    if port not in list(portnum.keys()):
        raise NameError('Invalid output port.')
    return portnum[port]

def get_outputbrakevalue(brake):

    """ Get the output brake value correposnding to the setting word """

    brakevalue = {
        'off': 0,
        'on': 1
    }
    brake = brake.lower()
    if type(brake) is not str:
        raise NameError('Invalid brake value.')
    if brake not in list(brakevalue.keys()):
        raise NameError('Invalid brake value.')
    return brakevalue[brake]



"""
Creates a class for interacting with the LEGO EV3 brick using direct commands.
LEGO original documentation on direct commands can be found in:
    - LEGO MINDSTORMS EV3 Firmware Developer Kit.pdf
    - LEGO MINDSTORMS EV3 Communication Developer Kit.pdf

Essential commands for motors and all commands for sensors are implemented.
File I/O and LCD commands are left for those with a brave spirit.

Note: Use code folding at the method level to improve readability.

Author: Eduardo Nigro
    rev 0.0.1
    2021-10-24
"""
import time
import numpy as np
from pyev3 import io
from pyev3 import comm


class LegoEV3:
    """
    The class to represent the LEGO EV3 brick
    You can use LegoEV3 to interact with the EV3 brick.

    Properties
    ----------
    devmode             - Flag (True or False) to display cmd/reply stats
    batterylevel        - EV3 battery level (0 to 100 %)
    connectedsensors    - List of connected sensors
    
    Methods
    -------
    close()             - Closes the current communication connection
    display_info()      - Displays summary info about the brick
    set_statuslight()   - Sets EV3 LED status light
    play_tone()         - Plays a tone on EV3 speaker

    Example
    -------
    Set up USB connection between host and EV3 brick.
    >>> from pyev3.brick import LegoEV3
    >>> myev3 = LegoEV3(commtype='usb')
    >>> myev3.display_info()
    >>> myev3.close()

    Set up WiFi connection between host and EV3 brick
    >>> myev3 = LegoEV3(commtype='wifi',
                        IPaddress='192.168.0.19, deviceID='001653470e58')
    >>> myev3.display_info()
    >>> myev3.close()

    Notes
    -----
    1. Always use the close() method before opening a new connection.
    2. Try a USB connection first. It's easier to set up and much faster.

    See also
    --------
    Motor, Touch, Color, Ultrasonic, Infrared, and Gyro

    """
    def __init__(self, commtype='usb', IPaddress=None, deviceID=None):
        """
        Class constructor.

        Parameters
        ----------
        commtype : str
            The type of communication with the brick.
            Default value is `usb`. Possible values are `usb` or `wifi`
        IPaddress : str
            The IP address assigned to the EV3 brick.
            Default value is None
        deviceID : str
            The individual device ID of the EV3 brick.
            Connect to the brick using commtype='usb' and use the
            display_info() method to retrieve the ID of the brick.
            Default value is None.

        Returns
        -------
        None.

        """
        # Assigning private property developer mode flag
        self._devmode = False
        # Assigning EV3 hardware number
        self.HARDWARE = 0x02
        # Assigning LED attributes
        self.LEDcolorlist = ['green', 'red', 'orange']
        self.LEDmodelist = ['solid', 'pulsing', 'off']
        self.LEDpattern = {
            'off': 0,
            'solidgreen': 1,
            'solidred': 2,
            'solidorange': 3,
            'pulsinggreen': 7,
            'pulsingred': 8,
            'pulsingorange': 9
        }
        # Assigning list of device (sensor) types
        self.devicetypes = {
            io.DeviceType.TYPE_EV3_TOUCH: 'EV3 Touch',
            io.DeviceType.TYPE_EV3_COLOR: 'EV3 Color',
            io.DeviceType.TYPE_EV3_ULTRASONIC: 'EV3 Ultrasonic',
            io.DeviceType.TYPE_EV3_GYRO: 'EV3 Gyro',
            io.DeviceType.TYPE_EV3_IR: 'EV3 Infrared',
            io.DeviceType.TYPE_THIRD_PARTY_START: 'Third Party Start',
            io.DeviceType.TYPE_THIRD_PARTY_END: 'Third Party End',
            io.DeviceType.TYPE_IIC_UNKNOWN: 'IIC Unknown',
            io.DeviceType.TYPE_TERMINAL: 'Terminal',
            io.DeviceType.TYPE_UNKNOWN: 'Unknown',
            io.DeviceType.TYPE_NONE: ' ',
            io.DeviceType.TYPE_ERROR: 'Error'
        }
        # Creating communication connection
        if commtype.lower() == 'usb':
            self._commtype = 'USB'
            self._commhandle = comm.CommUSB(deviceID)
        elif commtype.lower() == 'wifi':
            self._commtype = 'WiFi'
            self._commhandle = comm.CommTCPIP(IPaddress, deviceID)
        else:
            raise NameError('Unknown communication type.')
        # Getting EV3 properties
        self._brickname = self._get_brickname()
        self._brickID = self._get_brickid(self.HARDWARE)
        self._hardwareversion = self._read_hardwareversion()
        self._firmware = (
            'Version: ' + self._read_firmwareversion() +
            ' / Build: ' + self._read_firmwarebuild())
        self._os = (
            'Version: ' + self._read_osversion() +
            ' / Build: ' + self._read_osbuild())

    # GET/SET METHODS (PUBLIC PROPERTIES)
    @property
    def devmode(self):
        # Getting private value
        return self._devmode

    @devmode.setter
    def devmode(self, value):
        # Setting private vlaue
        self._devmode = value
        # Updating developer mode in communication object
        self._commhandle._devmode = self._devmode

    @property
    def batterylevel(self):
        return self._read_batterylevel()

    @batterylevel.setter
    def batterylevel(self, _):
        print('"batterylevel" is a read only attribute.')

    @property
    def connectedsensors(self):
        return self._read_inputdevicelist()

    @connectedsensors.setter
    def connectedsensors(self, _):
        print('"connectedsensors" is a read only attribute.')

    # BRICK PUBLIC METHODS
    def close(self):
        self._commhandle.close()

    def display_info(self):
        print('____________________________________________________________')
        print('Brick Name : ' + self._brickname)
        print('Brick ID : ' + self._brickID)
        print('Communication Type : ' + self._commtype)
        print('Hardware Version : ' + self._hardwareversion)
        print('Firmware : ' + self._firmware)
        print('OS : ' + self._os)
        print('Battery Level : ' + str(self.batterylevel))
        print('Connected Sensors : ' + str(self.connectedsensors))
        print('____________________________________________________________')

    def set_statuslight(self, mode='solid', color='green'):
        """
        Set the status light of the EV3 brick.

        Parameters
        ----------
        mode : str
            The light mode: `solid`, `pulsing`, `off`
            Default value is `solid`
        color : str
            The light color: `green`, `orange`, `red`
            Default value is `green`

        Returns
        -------
        None.

        Example
        -------
        >>> myev3.set_statuslight(mode='pulsing', color='orange')

        """
        pattern = self._get_LEDpattern(mode, color)
        cmd = io.DirectCommand.write_LED(pattern)
        self._commhandle.send(cmd)
        self._receive_reply()

    def play_tone(self, volume=10, frequency=440, duration=1):
        """
        Play a tone on the EV3 brick.

        Parameters
        ----------
        volume : int
            The volume of the played tone: 0 to 100
            Default value is 10
        frequency : int
            The frequency (Hz) of the played tone: 0 to 10000
            Default frequency is 440 Hz
        duration : float
            The duration (s) of the played tone: 0 to 5
            Default duration is 1.0 s

        Returns
        -------
        None.

        Example
        -------
        >>> myev3.play_tone(volume=5, frequency=880, duration=0.5)

        """
        cmd = io.DirectCommand.play_tone(volume, frequency, duration)
        self._commhandle.send(cmd)
        self._receive_reply()

    # BRICK PRIVATE METHODS
    def _get_brickid(self, hardware):
        # opCOM_GET, GET_ID
        cmd = io.DirectCommand.get_brickid(hardware)
        self._commhandle.send(cmd)
        return self._receive_reply(type='char')

    def _get_brickname(self):
        # opCOM_GET, GET_BRICKNAME
        cmd = io.DirectCommand.get_brickname()
        self._commhandle.send(cmd)
        return self._receive_reply(type='char')

    def _read_osversion(self):
        # opCOM_GET, GET_OS_VERS
        cmd = io.DirectCommand.read_osversion()
        self._commhandle.send(cmd)
        return self._receive_reply(type='char')

    def _read_hardwareversion(self):
        # opCOM_GET, GET_HW_VERS
        cmd = io.DirectCommand.read_hardwareversion()
        self._commhandle.send(cmd)
        return self._receive_reply(type='char')

    def _read_firmwareversion(self):
        # opCOM_GET, GET_FW_VERS
        cmd = io.DirectCommand.read_firmwareversion()
        self._commhandle.send(cmd)
        return self._receive_reply(type='char')

    def _read_firmwarebuild(self):
        # opCOM_GET, GET_FW_BUILD
        cmd = io.DirectCommand.read_firmwarebuild()
        self._commhandle.send(cmd)
        return self._receive_reply(type='char')

    def _read_osbuild(self):
        # opCOM_GET, GET_FW_BUILD
        cmd = io.DirectCommand.read_osbuild()
        self._commhandle.send(cmd)
        return self._receive_reply(type='char')

    def _read_batterylevel(self):
        # opCOM_GET, GET_FW_BUILD
        cmd = io.DirectCommand.read_batterylevel()
        self._commhandle.send(cmd)
        return self._receive_reply(type='int8')

    # MOTOR PRIVATE METHODS
    def _stop_motor(self, layer=1, port='A', brake='off'):
        # opOUTPUT_STOP
        cmd = io.DirectCommand.stop_motor(layer, port, brake)
        self._commhandle.send(cmd)
        self._receive_reply()

    def _set_outputpower(self, layer=1, port='A', power=10):
        # opOUTPUT_SET_TYPE (POWER)
        cmd = io.DirectCommand.set_outputpower(layer, port, power)
        self._commhandle.send(cmd)
        self._receive_reply()

    def _set_outputspeed(self, layer=1, port='A', speed=10):
        # opOUTPUT_SET_TYPE (SPEED)
        cmd = io.DirectCommand.set_outputspeed(layer, port, speed)
        self._commhandle.send(cmd)
        self._receive_reply()

    def _start_motor(self, layer=1, port='A'):
        # opOUTPUT_STOP
        cmd = io.DirectCommand.start_motor(layer, port)
        self._commhandle.send(cmd)
        self._receive_reply()

    def _clear_outputcount(self, layer=1, port='A'):
        # opOUTPUT_CLR_COUNT
        cmd = io.DirectCommand.clear_outputcount(layer, port)
        self._commhandle.send(cmd)
        self._receive_reply()

    def _get_outputcount(self, layer=1, port='A'):
        # opOUTPUT_GET_COUNT
        cmd = io.DirectCommand.get_outputcount(layer, port)
        self._commhandle.send(cmd)
        return self._receive_reply(type='int32')

    # SENSOR PRIVATE METHODS
    def _read_inputdevicelist(self):
        # opINPUT_DEVICE_LIST
        cmd = io.DirectCommand.read_inputdevicelist()
        data = self._send_cmd_receive_reply(cmd)
        return self._get_devciceinfo(data)

    def _read_inputdeviceformat(self, layer=1, portnum=1):
        # opINPUT_DEVICE, GET_FORMAT
        # [dataset, format, modes, view]
        cmd = io.DirectCommand.read_inputdeviceformat(layer, portnum)
        data = self._send_cmd_receive_reply(cmd)
        dataset = data[0]
        format = data[1]
        modes = data[2]
        view = data[3]
        return dataset, format, modes, view

    def _read_inputdevicetypemode(self, layer=1, portnum=1):
        # opINPUT_DEVICE, GET_TYPEMODE
        cmd = io.DirectCommand.read_inputdevicetypemode(layer, portnum)
        data = self._send_cmd_receive_reply(cmd, externalflag='type')
        type = data[0]
        mode = data[1]
        return type, mode

    def _read_inputdevicesymbol(self, layer=1, portnum=1):
        # opINPUT_DEVICE, GET_SYMBOL
        cmd = io.DirectCommand.read_inputdevicesymbol(layer, portnum)
        symbol = self._send_cmd_receive_reply(cmd, externalflag='string')
        return symbol

    def _read_inputdevicename(self, layer=1, portnum=1):
        # opINPUT_DEVICE, GET_NAME
        cmd = io.DirectCommand.read_inputdevicename(layer, portnum)
        name = self._send_cmd_receive_reply(cmd, externalflag='string')
        return name

    def _read_inputdevicemodename(self, layer=1, portnum=1, mode=0):
        # opINPUT_DEVICE, GET_MODENAME
        cmd = io.DirectCommand.read_inputdevicemodename(layer, portnum, mode)
        modename = self._send_cmd_receive_reply(cmd, externalflag='string')
        return modename

    def _clear_inputdevicechanges(self, layer=1, portnum=1):
        # opINPUT_DEVICE, CLR_CHANGES
        cmd = io.DirectCommand.clear_inputdevicechanges(layer, portnum)
        self._send_cmd_receive_reply(cmd)

    def _read_inputdevicereadySI(self, layer=1, portnum=1, mode=0, dataset=1):
        # opINPUT_DEVICE, READY_SI - Read sensor input in SI format
        cmd = io.DirectCommand.read_inputdevicereadySI(layer, portnum, mode, dataset)
        self._commhandle.send(cmd)
        data = self._receive_reply(type='other')
        return self._get_deviceoutput(data, dataset, type='SI')

    def _read_inputdevicereadyPCT(self, layer=1, portnum=1, mode=0, dataset=1):
        # opINPUT_DEVICE, READY_SI - Read sensor input in SI format
        cmd = io.DirectCommand.read_inputdevicereadyPCT(layer, portnum, mode, dataset)
        self._commhandle.send(cmd)
        data = self._receive_reply(type='other')
        return self._get_deviceoutput(data, dataset, type='PCT')

    def _read_inputdevicereadyRAW(self, layer=1, portnum=1, mode=0, dataset=1):
        # opINPUT_DEVICE, READY_SI - Read sensor input in SI format
        cmd = io.DirectCommand.read_inputdevicereadyRAW(layer, portnum, mode, dataset)
        self._commhandle.send(cmd)
        data = self._receive_reply(type='other')
        return self._get_deviceoutput(data, dataset, type='RAW')

    # HELPER METHODS
    def _receive_reply(self, type=None):
        #
        result = None
        data = self._commhandle.receive()
        # Checking for successful receive status
        if data[4] == 2:
            if type == 'char':
                result = "".join(map(chr, data[5:data.index(0, 5)]))
            elif type == 'int8':
                result = int(data[5])
            elif type == 'int32':
                if self._commtype == 'USB':
                    result = int(np.array(data[5:9], dtype=np.int8).view(np.int32))
                else:
                    result = int(np.array(data[5:9]).view(np.int32))
            elif type == 'other':
                result = data[5::]
            return result
        else:
            raise NameError('Get brick reply failed.')

    def _send_cmd_receive_reply(self, cmd, externalflag=None):
        attempt = 0
        dataflag = False
        while (attempt < 5) and not dataflag:
            time.sleep(0.1)
            self._commhandle.send(cmd)
            data = self._commhandle.receive()
            if data[4] == 2:
                if externalflag is None:
                    dataflag = True
                    result = data[5::]
                elif externalflag == 'type':
                    dataflag = self._is_validdevicetype(data)
                    result = data[5::]
                elif externalflag == 'string':
                    dataflag = self._is_validdevicestring(data)
                    result = "".join(map(chr, data[5:data.index(0, 5)])).strip()
            attempt += 1
        if dataflag:
            return result
        else:
            raise NameError('Get brick reply failed.')

    def _get_LEDpattern(self, mode, color):
        #
        if mode not in self.LEDmodelist:
            raise NameError('Invalid LED mode argument.')
        if color not in self.LEDcolorlist:
            raise NameError('Invalid LED color argument.')
        if mode == 'off':
            color = ''
        return self.LEDpattern[mode+color]

    def _get_devciceinfo(self, data):
        #
        info = []
        for type in data:
            info.append(self.devicetypes[type])
        return info

    def _get_deviceoutput(self, data, dataset, type='SI'):
        #
        result = []
        for i in range(dataset):
            if type == 'SI':
                if self._commtype == 'USB':
                    result.append(np.array(data[i*4:i*4+4], dtype=np.int8).view(np.float32)[0])
                else:
                    result.append(np.array(data[i*4:i*4+4]).view(np.float32))
            elif type == 'RAW':
                if self._commtype == 'USB':
                    result.append(np.array(data[i*4:i*4+4], dtype=np.int8).view(np.int32)[0])
                else:
                    result.append(np.array(data[i*4:i*4+4]).view(np.int32))
            elif type == 'PCT':
                result.append(data[i])
        return result

    def _is_validdevicetype(self, data):
        #
        type = data[5]
        if type in list(self.devicetypes.keys()):
            validtype = True
        else:
            validtype = False
        return validtype

    def _is_validdevicestring(self, data):
        #
        string = "".join(map(chr, data[0:data.index(0)])).strip()
        if len(string) > 0:
            validstring = True
        else:
            validstring = False
        return validstring



""" 
comm.py contains the classes used to communicate with the EV3 brick.
Classes that support USB and TCP/IP communication are implemented.

Author: Eduardo Nigro
    rev 0.0.1
    2021-10-24
"""
import hid
import socket
import struct
import time
import numpy as np


class CommUSB:
    def __init__(self, deviceID):

        """ Class constructor """ 

        # Defining USB constants
        ID_VENDOR_LEGO = 0x0694
        ID_PRODUCT_EV3 = 0x0005
        # Connecting to EV3
        self.usbhandle = hid.device()
        if not deviceID:
            self.usbhandle.open(ID_VENDOR_LEGO, ID_PRODUCT_EV3)
        else:
            self.usbhandle.open(ID_VENDOR_LEGO, ID_PRODUCT_EV3, deviceID)
        # Assigning developer mode flag
        self._devmode = False

    def send(self, cmd):

        """ Send direct command data """

        # Padding command with zeros for 1024-byte package
        data = bytes(1) + cmd + bytes(1023-len(cmd))
        # Collecting developer mode data
        if self._devmode:
            self.cmd = cmd
            self.tstart = time.perf_counter_ns()
        # Sending direct command to EV3 brick
        self.usbhandle.write(data)

    def receive(self):

        """ Receive direct command reply data """

        # Getting 1024-byte data package
        data = self.usbhandle.read(1024)
        # Extracting actual number of bytes in the reply
        size = np.array(data[0:2], dtype=np.uint8).view(np.uint16)
        data = data[0:size[0]+2]
        # Collecting developer mode data and displaying info
        if self._devmode:
            self.tend = time.perf_counter_ns()-self.tstart
            self.reply = data
            display_stats(self.cmd, self.reply, self.tend)
        return data

    def close(self):

        """ Closes EV3 TCP client """

        self.usbhandle.close()

class CommTCPIP:
    def __init__(self, IPaddress, deviceID):

        """ Class constructor """ 

        # Defining TCP/IP parameters
        self.TCP_PORT = 5555
        self.BUFFER_SIZE = 1024
        self.TCP_IP = IPaddress
        self.deviceID = deviceID
        # Creating TCP/IP socket
        self.tcphandle = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.tcphandle.settimeout(5)
        self.tcphandle.connect((self.TCP_IP, self.TCP_PORT))
        # Connecting to EV3
        unlockstr = 'GET /target?sn=' + deviceID + ' VMTP1.0\nProtocol: EV3'
        self.tcphandle.send(unlockstr.encode())
        data = self.tcphandle.recv(self.BUFFER_SIZE)
        if data.decode('utf-8').find('Accept:EV340') < 0:
            raise NameError('Incorrect device.')
        # Assigning developer mode flag
        self._devmode = False
        # Setting socket as blocking
        self.tcphandle.settimeout(None)

    def send(self, cmd):

        """ Send direct command data """

        # Collecting developer mode data
        if self._devmode:
            self.cmd = cmd
            self.tstart = time.perf_counter_ns()
        # Sending direct command to EV3 brick
        self.tcphandle.send(cmd)

    def receive(self):

        """ Receive direct command reply data """

        # Getting 1024-byte data package
        data = self.tcphandle.recv(self.BUFFER_SIZE)
        # Collecting developer mode data and displaying info
        if self._devmode:
            self.tend = time.perf_counter_ns()-self.tstart
            self.reply = data
            display_stats(self.cmd, self.reply, self.tend)
        return data

    def close(self):

        """ Closes EV3 TCP client """

        self.tcphandle.shutdown(socket.SHUT_RDWR)
        self.tcphandle.close()

# HELPER FUNCTIONS
def display_stats(cmd, reply, t):

    """ Displays cmd/reply values and round-trip elapsed time. """

    print('____________________________________________________________')
    print('cmd:')
    print('  uint8 ' + str([byte[0]
            for byte in list(struct.iter_unpack('B', cmd))]))
    print('  0x|' + ':'.join('{:02X}'.format(byte) for byte in cmd) + '|')
    print('reply:')
    print('  uint8 ' + str(reply))
    print('  0x|' + ':'.join('{:02X}'.format(byte) for byte in reply) + '|')
    print("cmd/reply elapsed time (ms): {:3.3f}".format(1e-6*t))
    print('____________________________________________________________')
    print('')

def list_bricks():
    """
    Return a dictionary containing the names and corresponding ID of
    all the EV3 bricks that are currently connected to the USB port.
    
    Parameters
    ----------
    None.

    Returns
    -------
    brick : dicitonary containing brick names and IDs.

    Example
    -------
    >>> mybricks = list_bricks()

    """
    ID_VENDOR_LEGO = 0x0694
    brick = {
        'Name': [],
        'ID': []
    }
    for device_dict in hid.enumerate():
        if device_dict['vendor_id'] == ID_VENDOR_LEGO:
            brick['Name'].append(device_dict['product_string'])
            brick['ID'].append(device_dict['serial_number'])
    return brick


""" 
devices.py contains the Motor and Sensor classes.
Each sensor type has its own class, derived from the base Sensor class.
Motor and sensor classes implement Direct Commands in a way that simplifies 
the interaction with the EV3 devices.

Author: Eduardo Nigro
    rev 0.0.1
    2021-10-24
"""
import numpy as np
import time

class Motor:
    """
    The class to represent the EV3 motor
    You can use Motor to interact with the EV3 motor.
    
    Properties
    ----------
    outputmode          - Motor output mode
    output              - Output level of the motor
    angle               - Current anglular position of the motor
    
    Methods
    -------
    start()             - Start motor
    stop()              - Stop motor
    reset_angle()       - Reset angular position to zero
    
    Example
    -------
    Set up a motor on output port 'B'
    >>> from pyev3.brick import LegoEV3
    >>> from pyev3.devices import Motor
    >>> myev3 = LegoEV3(commtype='usb')
    >>> mymotor = Motor(myev3, port='B')

    Notes
    -----
    1. If EV3 bricks are connected in a daisy-chain configuration,
       set the input parameter `layer` to the appropriate number.

    Set up a motor on output port 'A' of first brick
    >>> mymotor1 = Motor(ev3, layer=1, port='A')
    Set up a motor on output port 'A' of second brick
    >>> mymotor2 = Motor(ev3, layer=2, port='A')
    
    See also
    --------
    LegoEV3, Touch, Color, Ultrasonic, Infrared, and Gyro

    """
    def __init__(self, ev3handle, layer=1, port='A'):
        """
        Class constructor.

        Parameters
        ----------
        ev3handle : LegoEV3 instance
            The object representing the EV3 brick.
        layer : int
            The layer of the brick in a daisy-chain configuration.
            Default value is 1. Possible values are 1 or 2.
        port : str
            The brick output port connected to the motor.
            Default value is `A`. Possible values are `A`, `B`, `C`, or `D`.

        Returns
        -------
        None.

        """
        # Assigning input attributes
        self._ev3handle = ev3handle
        self._layer = layer
        self._outputport = port
        # Assigning default attributes
        self._motorstopped = True
        self._outputmode = 'power'
        self._output = 0

    # GET/SET METHODS (PUBLIC PROPERTIES)
    @property
    def outputmode(self):
        return self._outputmode

    @outputmode.setter
    def outputmode(self, value):
        if self._motorstopped:
            if value.lower() == 'speed':
                self._outputmode = 'speed'
            elif value.lower() == 'power':
                self._outputmode = 'power'
            else:
                raise NameError('Invalid output type.')
        else:
            print('Stop motor to change output mode.')

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, value):
        # Limiting value or changing data type
        if value > 100:
            value = 100
        elif value < -100:
            value = -100
        else:
            value = np.int8(value)
        # Sending appropriate output level command to EV3 brick
        if self.outputmode == 'power':
            self._ev3handle._set_outputpower(
                layer=self._layer, port=self._outputport, power=value)
        elif self.outputmode == 'speed':
            self._ev3handle._set_outputspeed(
                layer=self._layer, port=self._outputport, speed=value)
        # Updating Output property
        self._output = value

    @property
    def angle(self):
        return self._ev3handle._get_outputcount(
            layer=self._layer, port=self._outputport)

    @angle.setter
    def angle(self, _):
        print('"angle" is a read only property.')

    # MOTOR CONTROL METHODS
    def start(self):
        """
        Start the EV3 motor.

        Parameters
        ----------
        None.

        Returns
        -------
        None.

        Example
        -------
        >>> mymotor.start()

        """
        # Making sure current output mode is sent to EV3 brick
        if self.outputmode == 'power':
            self._ev3handle._set_outputpower(
                layer=self._layer, port=self._outputport, power=self.output)
        elif self.outputmode == 'speed':
            self._ev3handle._set_outputspeed(
                layer=self._layer, port=self._outputport, speed=self.output)
        # Changing motor stopped status flag
        self._motorstopped = False
        # Starting motor
        self._ev3handle._start_motor(
            layer=self._layer, port=self._outputport)

    def stop(self, brake='off'):
        """
        Stop the EV3 motor.

        Parameters
        ----------
        brake : str
            The brake option of the motor `on` or `off`.
            Can be used to hold the motor position.
            Default value is `off`

        Returns
        -------
        None.

        Example
        -------
        >>> mymotor.stop(brake='on')

        """
        # Changing motor stopped status flag
        self._motorstopped = True
        # Stopping motor with appropriate brake option
        self._ev3handle._stop_motor(
            layer=self._layer, port=self._outputport, brake=brake)

    def reset_angle(self):
        """
        Reset the encoder output angle of the EV3 motor.

        Parameters
        ----------
        None.

        Returns
        -------
        None.

        Example
        -------
        >>> mymotor.reset_angle()

        """
        self._ev3handle._clear_outputcount(
            layer=self._layer, port=self._outputport)


class Sensor:
    """
    The Sensor class implements common methods for all the different types
    of EV3 sensors. The following properties and methods are inherited by
    the derived sensor-specific classes.

    Properties
    ----------
    inputmode           - The sensor input mode (see each sensor for options)
    
    Methods
    -------
    display_info()      - Displays summary info about the sensor
    
    Notes
    -----
    1. The sensor input mode can be changed at any time,
       even after its creation.

    """
    def __init__(self, ev3handle, layer=1, portnum=1):
        """
        Class constructor.

        Parameters
        ----------
        ev3handle : LegoEV3 instance
            The object representing the EV3 brick.
        layer : int
            The layer of the brick in a daisy-chain configuration.
            Default value is 1. Possible values are 1 or 2.
        portnum : int
            The brick input port connected to the sensor.
            Default value is 1. Possible values are 1, 2, 3, or 4.

        Returns
        -------
        None.

        """
        # Assigning input attributes
        self._ev3handle = ev3handle
        self._layer = layer
        self._inputport = portnum
        # Assigning sensor specific attributes
        self._sensortype, self.activemode = self._ev3handle._read_inputdevicetypemode(
            layer=self._layer, portnum=self._inputport)
        # Getting sensor info and separating name from mode
        nameaux = self._ev3handle._read_inputdevicename(
            layer=self._layer, portnum=self._inputport)
        indexaux = nameaux.find('-')
        # Assigning sensor name and active mode attributes
        if indexaux < 0:
            self._sensorname = 'TOUCH'
            self._activemodename = nameaux
        else:
            self._sensorname = nameaux[0:indexaux]
            self._activemodename = nameaux[indexaux+1::]
        # Creating default generic sensor attributes
        self._inputmode = None
        self._outputformat = None
        self._mode = None
        self._dataset = None
        # Creating sensor info dictioanry
        self._info = {
            'TOUCH': ['Touch', ['touch', 'bump']],
            'COL': ['Color', ['reflected', 'ambient', 'color', 'rgb']],
            'RGB': ['Color', ['reflected', 'ambient', 'color', 'rgb']],
            'US': ['Ultrasonic', ['distance', 'listen']],
            'IR': ['Infrared', ['proximity', 'seeker', 'remote']],
            'GYRO': ['Gyro', ['angle', 'rate', 'angle&rate']]
        }

    # GET/SET METHOD INPUTMODE
    @property
    def inputmode(self):
        return self._inputmode

    @inputmode.setter
    def inputmode(self, value):
        if value in list(self.modepar.keys()):
            self._inputmode = value
            self._outputformat = self.modepar[value][0]
            self._mode = self.modepar[value][1]
            self._dataset = self.modepar[value][2]
            # Forcing mode change
            _ = self._read_output()
        else:
            nameaux = self._info[self._sensorname][0]
            raise NameError('Undefined input mode for ' + nameaux + ' sensor.')

    # SENSOR BASE CLASS METHODS
    def display_info(self):
        print('____________________________________________________________')
        print('Sensor Name : ' + self._info[self._sensorname][0])
        print('Current Mode : ' + self.inputmode)
        print('Available Modes : ' + str(self._info[self._sensorname][1])[1:-1])
        print('Layer : ' + str(self._layer))
        print('Port Number : ' + str(self._inputport))
        print('____________________________________________________________')

    def _read_output(self):
        
        """ Read generic sensor output value. """

        # Assigning sensor output reader method
        if self._outputformat == 1:
            reader = self._ev3handle._read_inputdevicereadySI
        elif self._outputformat == 2:
            reader = self._ev3handle._read_inputdevicereadyPCT
        elif self._outputformat == 3:
            reader = self._ev3handle._read_inputdevicereadyRAW
        # Reading sensor output
        result = reader(
            layer=self._layer, portnum=self._inputport,
            mode=self._mode, dataset=self._dataset)
        while result is None:
            time.sleep(0.01)
            result = reader(
                layer=self._layer, portnum=self._inputport,
                mode=self._mode, dataset=self._dataset)
        return result

    def _clear_changes(self):
        self._ev3handle._clear_inputdevicechanges(
            layer=self._layer, portnum=self._inputport)


class Touch(Sensor):
    """
    The class to represent the EV3 touch sensor.
    
    Properties
    ----------
    inputmode
    - 'touch'
    - 'bump'
    output
    - 1 or 0  (pressed or released button)
    - integer (press/release event count)
    
    Methods
    -------
    reset_count()       - Reset 'bump' counter
    
    Example
    -------
    Set up a touch sensor on port number 1
    >>> from pyev3.brick import LegoEV3
    >>> from pyev3.devices import Touch
    >>> myev3 = LegoEV3(commtype='usb')
    >>> mysensor = Touch(myev3, portnum=1, inputmode='bump')

    See also
    --------
    LegoEV3, Motor, Color, Ultrasonic, Infrared, and Gyro

    """
    def __init__(self, ev3handle, layer=1, portnum=1, inputmode=None):
       
        """ Class constructor """

        # Defining touch sensor mode parameters
        # [outputformat, mode, dataset]
        self.modepar = {
            'touch': [1, 0, 1],
            'bump': [1, 1, 1]
        }
        # Instantiating base class Sensor
        super(Touch, self).__init__(ev3handle, layer=layer, portnum=portnum)
        # Checking for valid sensor
        if self._info[self._sensorname][0] != 'Touch':
            raise NameError('Incorrect sensor type on port : ' + str(portnum))
        # Assining sensor specific attributes
        self.inputmode = inputmode

    @property
    def output(self):
        data = self._read_output()
        result = int(data[0])
        return result

    @output.setter
    def output(self, _):
        print('"output" is a read only property.')

    def reset_count(self):
        return self._clear_changes()


class Color(Sensor):
    """
    The class to represent the EV3 color sensor.
    
    Properties
    ----------
    inputmode
    - 'ambient'
    - 'reflected'
    - 'color'
    - 'rgb'
    output
    - 0 to 100 (ambient light intensity)
    - 0 to 100 (reflected light intensity)
    - 'black', 'blue', 'green', 'yellow', 'red', 'white', 'brown', 'unknown'
    - tuple of (R, G, B) integer components between 0 and 255
    
    Example
    -------
    Set up a color sensor on port number 2
    >>> from pyev3.brick import LegoEV3
    >>> from pyev3.devices import Color
    >>> myev3 = LegoEV3(commtype='usb')
    >>> mysensor = Color(myev3, portnum=2, inputmode='rgb')

    See also
    --------
    LegoEV3, Motor, Touch, Ultrasonic, Infrared, and Gyro

    """
    def __init__(self, ev3handle, layer=1, portnum=1, inputmode=None):
       
        """ Class constructor """

        # Defining color sensor mode parameters
        # [outputformat, mode, dataset]
        self.modepar = {
            'reflected': [2, 0, 1],
            'ambient': [2, 1, 1],
            'color': [1, 2, 1],
            'rgb': [3, 4, 3]
        }
        # Instantiating base class Sensor
        super(Color, self).__init__(ev3handle, layer=layer, portnum=portnum)
        # Checking for valid sensor
        if self._info[self._sensorname][0] != 'Color':
            raise NameError('Incorrect sensor type on port : ' + str(portnum))
        # Assining sensor specific attributes
        self.inputmode = inputmode
        # Defining color list
        self._color = [
            'unknown',
            'black',
            'blue',
            'green',
            'yellow',
            'red',
            'white',
            'brown'
        ]

    @property
    def output(self):
        data = self._read_output()
        if self._inputmode == 'reflected':
            result = int(data[0])
        elif self._inputmode == 'ambient':
            result = int(data[0])
        elif self._inputmode == 'color':
            result = self._color[int(data[0])]
        elif self._inputmode == 'rgb':
            result = tuple([min(255, int(value)) for value in data])
        return result

    @output.setter
    def output(self, _):
        print('"output" is a read only property.')


class Ultrasonic(Sensor):
    """
    The class to represent the EV3 ultrasonic sensor.
    
    Properties
    ----------
    inputmode
    - 'distance'
    - 'listen'
    output
    - 0 to 250 (distance in cm)
    - 0 or 1   (ultrasound presence)
    
    Example
    -------
    Set up an ultrasonic sensor on port number 3
    >>> from pyev3.brick import LegoEV3
    >>> from pyev3.devices import Ultrasonic
    >>> myev3 = LegoEV3(commtype='usb')
    >>> mysensor = Ultrasonic(myev3, portnum=3, inputmode='distance')

    See also
    --------
    LegoEV3, Motor, Touch, Color, Infrared, and Gyro

    """
    def __init__(self, ev3handle, layer=1, portnum=1, inputmode=None):
       
        """ Class constructor """

        # Defining ultrasonic sensor mode parameters
        # [outputformat, mode, dataset]
        self.modepar = {
            'distance': [1, 0, 1],
            'listen': [1, 2, 1]
        }
        # Instantiating base class Sensor
        super(Ultrasonic, self).__init__(ev3handle, layer=layer, portnum=portnum)
        # Checking for valid sensor
        if self._info[self._sensorname][0] != 'Ultrasonic':
            raise NameError('Incorrect sensor type on port : ' + str(portnum))
        # Assining sensor specific attributes
        self.inputmode = inputmode

    @property
    def output(self):
        data = self._read_output()
        if self._inputmode == 'distance':
            result = np.float32(data[0])
        elif self._inputmode == 'listen':
            result = int(data[0])
        return result

    @output.setter
    def output(self, _):
        print('"output" is a read only property.')


class Infrared(Sensor):
    """
    The class to represent the EV3 infrared sensor.
    
    Properties
    ----------
    inputmode
    - 'proximity'
    - 'seeker' (requires channel 1 and beacon on)
    - 'remote'
    output
    - (proximity) : 0 to 100 (proximity to object)
    - (seeker)    : tuple of integers (azymuth, proximity)
                        azymuth   = -20 to 20
                        proximity = 0 to 100
    - (remote)    : tuple of integers (channel, buttoncode)
                        channel 1, 2, 3, 4
                        buttoncode
                        0  = No button
                        1  = Button 1
                        2  = Button 2
                        3  = Button 3
                        4  = Button 4
                        5  = Buttons 1 and 3
                        6  = Buttons 1 and 4
                        7  = Buttons 2 and 3
                        8  = Buttons 2 and 4
                        9  = Beacon Mode is on
                        10 = Buttons 1 and 2
                        11 = Buttons 3 and 4
    
    Example
    -------
    Set up an infrared sensor on port number 4
    >>> from pyev3.brick import LegoEV3
    >>> from pyev3.devices import Infrared
    >>> myev3 = LegoEV3(commtype='usb')
    >>> mysensor = Infrared(myev3, portnum=4, inputmode='remote')

    See also
    --------
    LegoEV3, Motor, Touch, Color, Ultrasonic, and Gyro

    """
    def __init__(self, ev3handle, layer=1, portnum=1, inputmode=None):
       
        """ Class constructor """

        # Defining infrared sensor mode parameters
        # [outputformat, mode, dataset]
        self.modepar = {
            'proximity': [1, 0, 1],
            'seeker': [1, 1, 2],
            'remote': [1, 2, 4]
        }
        # Instantiating base class Sensor
        super(Infrared, self).__init__(ev3handle, layer=layer, portnum=portnum)
        # Checking for valid sensor
        if self._info[self._sensorname][0] != 'Infrared':
            raise NameError('Incorrect sensor type on port : ' + str(portnum))
        # Assining sensor specific attributes
        self.inputmode = inputmode

    @property
    def output(self):
        data = self._read_output()
        if self._inputmode == 'proximity':
            result = int(data[0])
        elif self._inputmode == 'seeker':
            if not np.isnan(data[1]):
                result = (int(data[0]), int(data[1]))
            else:
                raise NameError('Incorrect remote channel or beacon is off.')
        elif self._inputmode == 'remote':
            index = np.nonzero(data)[0]
            if len(index) > 0:
                channel = index[0]+1
                button = int(data[index[0]])
                result = (channel, button)
            else:
                result = (0, 0)
        return result

    @output.setter
    def output(self, _):
        print('"output" is a read only property.')


class Gyro(Sensor):
    """
    The class to represent the EV3 gyro sensor.
    
    Properties
    ----------
    inputmode
    - 'angle'             - integer (angular position in deg)
    - 'rate'              - integer (angular speed up to 440deg/s)
    - 'angle&rate'        - integers (rate, angle)
    output
    - integer (angular position in deg)
    - integer (angular speed up to 440deg/s)
    - tuple of integers (rate, angle)
    
    Example
    -------
    Set up a gyro sensor on port number 1
    >>> from pyev3.brick import LegoEV3
    >>> from pyev3.devices import Gyro
    >>> myev3 = LegoEV3(commtype='usb')
    >>> mysensor = Gyro(myev3, portnum=1, inputmode='rate')

    See also
    --------
    LegoEV3, Motor, Touch, Color, Ultrasonic, and Infrared

    """
    def __init__(self, ev3handle, layer=1, portnum=1, inputmode=None):
       
        """ Class constructor """

        # Defining gyro sensor mode parameters
        # [outputformat, mode, dataset]
        self.modepar = {
            'angle': [1, 0, 1],
            'rate': [1, 1, 1],
            'angle&rate': [1, 3, 2]
        }
        # Instantiating base class Sensor
        super(Gyro, self).__init__(ev3handle, layer=layer, portnum=portnum)
        # Checking for valid sensor
        if self._info[self._sensorname][0] != 'Gyro':
            raise NameError('Incorrect sensor type on port : ' + str(portnum))
        # Assining sensor specific attributes
        self.inputmode = inputmode

    @property
    def output(self):
        data = self._read_output()
        if self._inputmode == 'angle':
            result = int(data[0])
        elif self._inputmode == 'rate':
            result = int(data[0])
        elif self._inputmode == 'angle&rate':
            result = (int(data[0]), int(data[1]))
        return result

    @output.setter
    def output(self, _):
        print('"output" is a read only property.')
